//
//  PTBitmapDataInfo.h
//  PTImageIO
//
//  Created by Rimas Mickevičius on 2013-08-29.
//
//

#import <Foundation/Foundation.h>

#import <Accelerate/Accelerate.h>

/**
 *
 *                  +- for future use
 *                  |
 *                  |                  +- data is flipped - image origin (left top) is at buffers "left bottom". Meanwhile buffers begin represents image left bottom pixel
 *                  |                  |
 *                  |                  |                     +- data components ordering info
 *                  |                  |                     |
 *                  |                  |                     |               +- data uses float values if (format & kPTBitmapDataFormatUsesFloatComponents) != 0
 *                  |                  |                     |               |
 * +----------------+--------------+  ++-+             +-----+-----+        ++-+
 * |                               |  |  |             |           |        |  |
 * +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+
 * |31|30|29|28|27|26|25|24| |23|22|21|20|19|18|17|16| |15|14|13|12|11|10|09|08| |07|06|05|04|03|02|01|00|
 * +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+ +--+--+--+--+--+--+--+--+
 *                                 |  |  |           |             |        |    |                       |
 *                                 ++-+  +-----+-----+             +---+----+    +-----------+-----------+
 *                                  |          |                       |                     |
 *                                  |          |                       |                     +- bits per component value = (format & kPTBitmapDataFormatBitsPerComponentInfoMask)
 *                                  |          |                       |
 *                                  |          |                       +- format endian info. Tells how to interpret data values when bits per component > 8. Basic formats does not include endian info = you need to set it by yourself
 *                                  |          |
 *                                  |          +- number of components value (0-15) shifted by 16 bits
 *                                  |
 *                                  +- data is unpremultiplied if this is equal to 1 (default value is 0 which means that data is premultiplied). If format does not have alpha channel, value is undefined.
 */


typedef NS_ENUM(uint32_t, PTBitmapDataEndian) {
    kPTBitmapDataUndefinedEndian  = (0 << 9),
    kPTBitmapDataLittleEndian     = (1 << 9),
    kPTBitmapDataBigEndian        = (2 << 9),
};

typedef NS_ENUM(uint32_t, PTBitmapDataComponentsOrdering) {
    kPTBitmapDataOrdering_Undefined                 = ( 0 << 12 | 0 << 16),
    kPTBitmapDataOrdering_BGRA                      = ( 1 << 12 | 4 << 16),
    kPTBitmapDataOrdering_RGBA                      = ( 2 << 12 | 4 << 16),
    kPTBitmapDataOrdering_ARGB                      = ( 3 << 12 | 4 << 16),
    kPTBitmapDataOrdering_RGB                       = ( 4 << 12 | 3 << 16),
    kPTBitmapDataOrdering_R                         = ( 5 << 12 | 1 << 16),
    kPTBitmapDataOrdering_RA                        = ( 6 << 12 | 2 << 16),
    kPTBitmapDataOrdering_AR                        = ( 7 << 12 | 2 << 16),
    kPTBitmapDataOrdering_CMYK                      = ( 8 << 12 | 4 << 16),
    kPTBitmapDataOrdering_CMYKA                     = ( 9 << 12 | 5 << 16),
    kPTBitmapDataOrdering_ACMYK                     = (10 << 12 | 5 << 16),
};

typedef NS_ENUM(uint32_t, PTBitmapDataFormat) {
    
    kPTBitmapDataFormatBitsPerComponentInfoMask     = 0xFF,         // no shift
    
    kPTBitmapDataFormatUsesFloatComponents          = (1 << 8),
    
    kPTBitmapDataFormatEndianInfoMask               = 0xE00,        // 9 shift
    
    kPTBitmapDataFormatOrderingInfoMask             = 0xFF000,      // 12 shift
    
    kPTBitmapDataFormatNumberOfComponentsInfoMask   = 0xF0000,      // 16 shift
    
    kPTBitmapDataFormatWithoutEndianInfoMask        = 0xFFFFF1FF,
    
    kPTBitmapDataFormatContentIsFlippedMask         = (1 << 20),    // if this is equal to 1, image data is beeing flipped. That means that image origin (left top) point is stored at buffers left bottom (buffer begins with image left bottom point)
    
    kPTBitmapDataFormatContentIsUnpremultipliedMask = (1 << 21),    // if this is equal to 1, image data is beeing unpremultiplied.
    
    // default formats
    kPTBitmapDataFormat_Undefined                   = 0,
    
    kPTBitmapDataFormat_UInt8_BGRA                  = ( 8 | kPTBitmapDataOrdering_BGRA),
    kPTBitmapDataFormat_UInt8_RGBA                  = ( 8 | kPTBitmapDataOrdering_RGBA),
    kPTBitmapDataFormat_UInt16_BGRA                 = (16 | kPTBitmapDataOrdering_BGRA),
    kPTBitmapDataFormat_UInt16_RGBA                 = (16 | kPTBitmapDataOrdering_RGBA),
    
    kPTBitmapDataFormat_UInt8_R                     = ( 8 | kPTBitmapDataOrdering_R),
    kPTBitmapDataFormat_UInt8_RA                    = ( 8 | kPTBitmapDataOrdering_RA),
    kPTBitmapDataFormat_UInt8_AR                    = ( 8 | kPTBitmapDataOrdering_AR),
    kPTBitmapDataFormat_UInt8_RGB                   = ( 8 | kPTBitmapDataOrdering_RGB),
    kPTBitmapDataFormat_UInt8_ARGB                  = ( 8 | kPTBitmapDataOrdering_ARGB),
    kPTBitmapDataFormat_UInt8_CMYK                  = ( 8 | kPTBitmapDataOrdering_CMYK),
    kPTBitmapDataFormat_UInt8_CMYKA                 = ( 8 | kPTBitmapDataOrdering_CMYKA),
    kPTBitmapDataFormat_UInt8_ACMYK                 = ( 8 | kPTBitmapDataOrdering_ACMYK),
    kPTBitmapDataFormat_UInt16_R                    = (16 | kPTBitmapDataOrdering_R),
    kPTBitmapDataFormat_UInt16_RA                   = (16 | kPTBitmapDataOrdering_RA),
    kPTBitmapDataFormat_UInt16_AR                   = (16 | kPTBitmapDataOrdering_AR),
    kPTBitmapDataFormat_UInt16_RGB                  = (16 | kPTBitmapDataOrdering_RGB),
    kPTBitmapDataFormat_UInt16_ARGB                 = (16 | kPTBitmapDataOrdering_ARGB),
    kPTBitmapDataFormat_UInt16_CMYK                 = (16 | kPTBitmapDataOrdering_CMYK),
    kPTBitmapDataFormat_UInt16_CMYKA                = (16 | kPTBitmapDataOrdering_CMYKA),
    kPTBitmapDataFormat_UInt16_ACMYK                = (16 | kPTBitmapDataOrdering_ACMYK),
    
    kPTBitmapDataFormat_Float16_BGRA                = (16 | kPTBitmapDataOrdering_BGRA | kPTBitmapDataFormatUsesFloatComponents),
    kPTBitmapDataFormat_Float16_RGBA                = (16 | kPTBitmapDataOrdering_RGBA | kPTBitmapDataFormatUsesFloatComponents),
    kPTBitmapDataFormat_Float32_BGRA                = (32 | kPTBitmapDataOrdering_BGRA | kPTBitmapDataFormatUsesFloatComponents),
    kPTBitmapDataFormat_Float32_RGBA                = (32 | kPTBitmapDataOrdering_RGBA | kPTBitmapDataFormatUsesFloatComponents),
};

typedef struct {
    PTBitmapDataFormat  dataFormat;
    vImage_Buffer       dataBuffer;
    BOOL                forceBufferFreeingOnCleanup;
} PTBitmapDataBuffer;


/**
 * Defines what color space mode represents format
 */
typedef NS_ENUM(uint8_t, PTBitmapDataFormatColorMode)
{
    kPTBitmapDataFormatUndefinedColorMode    = 0,
    kPTBitmapDataFormatGrayColorMode         = 1,
    kPTBitmapDataFormatRGBColorMode          = 2,
    kPTBitmapDataFormatCMYKColorMode         = 3,
};

/**
 * returns bits count used to store one component data
 */
uint8_t PTBitmapDataFormatGetBitsPerComponent(PTBitmapDataFormat format);

/**
 * returns bytes count used to store one pixel data
 */
uint8_t PTBitmapDataFormatGetBytesPerPixel(PTBitmapDataFormat format);

/**
 * returns if format uses float values to store data
 */
 BOOL PTBitmapDataFormatUsesFloatValues(PTBitmapDataFormat format);

/**
 * returns number of components
 */
uint8_t PTBitmapDataFormatGetNumberOfComponents(PTBitmapDataFormat format);

/**
 * returns format data endian
 */
PTBitmapDataEndian PTBitmapDataFormatGetEndiant(PTBitmapDataFormat format);

/**
 * returns system endian
 */
PTBitmapDataEndian PTBitmapDataSystemEndian();

/**
 * returns format data components ordering
 */
PTBitmapDataComponentsOrdering PTBitmapDataFormatGetOrdering(PTBitmapDataFormat format);

extern uint8_t const kPTBitmapDataFormatUndefinedAlphaIndex;

/**
 * Returns the index of the alpha component in the pixel.
 * If format is undefined (kPTBitmapDataFormat_Undefined) or does not have alpha component, returns kPTBitmapDataFormatUndefinedAlphaIndex. You should always check if returned value is valid.
 */
uint8_t PTBitmapDataFormatGetAlphaIndex(PTBitmapDataFormat format);

/**
 * Returns CGBitmapInfo constructed from format
 */
CGBitmapInfo PTBitmapDataFormatGetCGBitmapInfo(PTBitmapDataFormat format);


/**
 * Creates and returns vImage_CGImageFormat from PTBitmapDataFormat with desired colorspace.
 */
vImage_CGImageFormat PTBitmapDataFormatGetCGImageFormat(PTBitmapDataFormat format, CGColorSpaceRef colorspace);

/**
 * Returns if format data is flipped.
 * All default bitmap data formats are not flipped. This means that image origin (left top pixel) is beeing stored at the beggining of the buffer.
 * Meanwhile flipped format tells, that buffer starts with left bottom pixel of the image.
 */
BOOL PTBitmapDataFormatIsFlipped(PTBitmapDataFormat format);

/**
 * Returns if format data is unpremultiplied.
 * If format does not have alpha component, returns YES
 */
BOOL PTBitmapDataFormatIsUnpremultiplied(PTBitmapDataFormat format);

/**
 * Returns color mode, which is beeing represented by supplied format
 */
PTBitmapDataFormatColorMode PTBitmapDataFormatGetColorMode(PTBitmapDataFormat format);



PTBitmapDataFormat PTBitmapDataFormatWithoutEndianInfo(PTBitmapDataFormat format);

PTBitmapDataFormat PTBitmapDataFormatWithoutFlippingInfo(PTBitmapDataFormat format);

PTBitmapDataFormat PTBitmapDataFormatWithoutUnpremultiplyInfo(PTBitmapDataFormat format);

/**
 * Strips optional format info flags (currently endian, flipping and unpremultiply) and returns one from default formats
 * This is usefull when format is needed to compare with default constants - for example if (PTBitmapDataDefaultFormatInfo(myFormat) == kPTBitmapDataFormat_UInt8_BGRA)
 */
PTBitmapDataFormat PTBitmapDataDefaultFormatInfo(PTBitmapDataFormat format);

PTBitmapDataFormat PTBitmapDataFormatConstruct(uint8_t bitsPerComponent, PTBitmapDataComponentsOrdering componentsOrdering, BOOL usesFloats, PTBitmapDataEndian endian, BOOL flipped, BOOL isUnpremultiplied);

PTBitmapDataFormat PTBitmapDataFormatBySettingEndian(PTBitmapDataFormat format, PTBitmapDataEndian newEndian);

/**
 *  Return format which is identical to the provided one except that system endian is being applied.
 */
PTBitmapDataFormat PTBitmapDataFormatBySettingSystemEndian(PTBitmapDataFormat format);

PTBitmapDataFormat PTBitmapDataFormatBySettingBitsPerComponents(PTBitmapDataFormat format, uint8_t newBitsPerComponentValue);

PTBitmapDataFormat PTBitmapDataFormatBySettingComponentsOrdering(PTBitmapDataFormat format, PTBitmapDataComponentsOrdering newComponentsOrdering);

PTBitmapDataFormat PTBitmapDataFormatBySettingFlip(PTBitmapDataFormat format, BOOL isFlipped);

PTBitmapDataFormat PTBitmapDataFormatBySettingUnpremultiplication(PTBitmapDataFormat format, BOOL isUnpremultiplied);

/*
PTBitmapDataFormat PTBitmapdataFormatByRemovingAlphaComponent(PTBitmapDataFormat formatToRemoveAlphaFrom, BOOL *success);

PTBitmapDataFormat PTBitmapdataFormatByAddingAlphaComponent(PTBitmapDataFormat formatToAddAlphaTo, BOOL *success);
 */


size_t PTBitmapDataBufferRowBytesForWidthUsingFormat(size_t bufferWidth, PTBitmapDataFormat format, BOOL shouldAlign);

/**
 *  This function is a convenience method to help initialize a PTBitmapDataBuffer struct with a buffer sized
 *  and aligned according to the format. It will initialize the format, height, width and rowBytes fields, and allocate
 *  the pixel storage for you. You are responsible for releasing the memory pointed to by bufferRefToCreate->dataBuffer.data back to
 *  the system when you are done with it using free() (you can use a helper function PTBitmapDataBufferCleanup to do that). If no such allocation is desired, pass
 *  NO in the allocateDataBuffer to cause bufferRefToCreate->dataBuffer.data to be set to NULL.
 *  Pass NO in alignRowBytes to ignore buffer rowbytes alignment. 
 */
void PTBitmapDataBufferInit(PTBitmapDataBuffer *bufferRefToInit, PTBitmapDataFormat format, size_t width, size_t height, BOOL alignRowBytes, BOOL allocateDataBuffer);

void PTBitmapDataBufferInitWithCustomData(PTBitmapDataBuffer *bufferRefToInit, PTBitmapDataFormat format, size_t width, size_t height, size_t rowBytes, void *data);

/**
 * Sets default values for provided bitmap data buffer
 */
void PTBitmapDataBufferSetDefaults(PTBitmapDataBuffer *bufferToInitiateRef);

/**
 * Sets provided buffer values to default ones (same as PTBitmapDataBufferSetDefaults. If releaseData == YES and buffer contains non NULL data, it is beeing freed. 
 * Returns BOOL to indicate if actual data freeing was performed.
 */
BOOL PTBitmapDataBufferCleanup(PTBitmapDataBuffer *bufferToCleanRef, BOOL releaseData);

/**
 *  Copies the data from the src data buffer into the dest data buffer.
 *  Parameters alignRowBytes and allocateDataBuffer are the same as for PTBitmapDataBufferInit function.
 */
void PTBitmapDataBufferCopy(PTBitmapDataBuffer *dest, PTBitmapDataBuffer *src, BOOL alignRowBytes, BOOL allocateDataBuffer);

/**
 * Returns data buffer length (rowbytes * height)
 */
NSUInteger PTBitmapDataBufferGetDataLength(PTBitmapDataBuffer *bufferRef);

/**
 * Premultiplies buffer data. If data format does not have alpha or buffer is already premultiplied, no operation is beeing performed.
 * Returns YES if actual premultiplication was performed, NO otherwise.
 */
BOOL PTBitmapDataBufferPremultiply(PTBitmapDataBuffer *bufferRef);

/**
 * Unpremultiplies buffer data. If data format does not have alpha or buffer is unpremultiplied, no operation is beeing performed.
 * Returns YES if actual unpremultiplication was performed, NO otherwise.
 */
BOOL PTBitmapDataBufferUnpremultiply(PTBitmapDataBuffer *bufferRef);


void PTBitmapDataFormatPrintInfo(PTBitmapDataFormat format, NSString *title);

/**
 * Logs information of provided buffer
 */
void PTBitmapDataBufferPrintInfo(PTBitmapDataBuffer buffer, NSString *title);

/**
 * Prints information bout all predefined formats
 */
void debug_PTBitmapDataFormatPrintAllFormatsInfo();








BOOL PTBitmapDataBufferFillChannel(PTBitmapDataBuffer *bufferRef, uint8_t channelIndex, uint32_t value);






@interface PTBitmapDataBufferWrapper : NSObject

- (PTBitmapDataBuffer *)bitmapDataBufferRef;

@property BOOL freeBufferUponDealloc;

@end



